/* HPCU Bi-Weekly Challenge: Hybrid Parallel SIR
 * Author: Aaron Weeden, Shodor, 2014
 */

/* import libraries */
#include <omp.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

/* create macro for exiting if MPI errors occur */
#define tryMPI(ret, str, rank) \
    if (ret != MPI_SUCCESS) { \
        fprintf(stderr, "Rank %d returned error for %s, exiting!\n", \
                rank, str); \
        exit(EXIT_FAILURE); \
    }

/* declare constants */
const int INIT_NUM_INFECTEDS = 1;
const int INIT_NUM_RECOVEREDS = 0;
const float INFECTION_FRACTION = 0.00218;
const int NUM_TIME_STEPS = 50;
const int MPI_TAG = 0;

/* start program */
int main(int argc, char **argv) {

    /* declare variables */
    int mpi_rank;
    int mpi_size;
    int num_threads;
    int init_num_susceptibles;
    float recovery_fraction;
    int num_susceptibles;
    int num_infecteds;
    int num_recovereds;
    int get_sick;
    int get_better;
    int i;
    int j;
    int *final_num_recovereds;

    /* set up mpi */
    tryMPI(
            MPI_Init(&argc, &argv),
            "MPI_Init",
            0
          );
    tryMPI(
            MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank),
            "MPI_Comm_rank",
            0
          );
    tryMPI(
            MPI_Comm_size(MPI_COMM_WORLD, &mpi_size),
            "MPI_Comm_size",
            mpi_rank
          );

    /* get number of threads */
#pragma omp parallel
    {
#pragma omp single
        {
            num_threads = omp_get_num_threads();
        }
    }

    /* allocate memory for array (enough space for each thread) */
    if ((final_num_recovereds = malloc(num_threads * sizeof(int))) == NULL) {
        fprintf(stderr, "Rank %d returned error for malloc, exiting!\n",
                mpi_rank);
        exit(EXIT_FAILURE);
    }

    /* skip for rank 0 */
    if (mpi_rank != 0) {

        /* parameterize initial number of susceptibles */
        init_num_susceptibles = 1000 * mpi_rank;

        /* spawn OpenMP threads */
#pragma omp parallel private(recovery_fraction, num_susceptibles, \
        num_infecteds, num_recovereds, get_sick, get_better, i)
        {
            /* parameterize recovery fraction */
            recovery_fraction = (float)(omp_get_thread_num()+1) /
                (num_threads+1);

            /* run a simulation */
            num_susceptibles = init_num_susceptibles;
            num_infecteds = INIT_NUM_INFECTEDS;
            num_recovereds = INIT_NUM_RECOVEREDS;
            for (i = 0; i < NUM_TIME_STEPS; i++) {

                /* calculate how many people get sick */
                get_sick = INFECTION_FRACTION * num_susceptibles *
                    num_infecteds;

                /* make sure we aren't making more people sick than there are
                 * people to make sick */
                if (get_sick > num_susceptibles)
                    get_sick = num_susceptibles;

                /* calculate how many people get better */
                get_better = recovery_fraction * num_infecteds;

                /* update the people counts */
                num_susceptibles -= get_sick;
                num_infecteds += get_sick - get_better;
                num_recovereds += get_better;
            }

            /* store the final number of recovereds in an array */
            final_num_recovereds[omp_get_thread_num()] = num_recovereds;

            /* done spawning threads */
        }

        /* send array to rank 0 */
        tryMPI(
                MPI_Send(final_num_recovereds, num_threads, MPI_INT, 0,
                    MPI_TAG, MPI_COMM_WORLD),
                "MPI_Send",
                mpi_rank
              );

        /* start rank 0, end other processes */
    } else {

        /* receive array from each other process */
        for (i = 1; i < mpi_size; i++) {
            tryMPI(
                    MPI_Recv(final_num_recovereds, num_threads, MPI_INT, i,
                        MPI_TAG, MPI_COMM_WORLD, MPI_STATUS_IGNORE),
                    "MPI_Recv",
                    mpi_rank
                  );

            /* print results */
            for (j = 0; j < num_threads; j++) {
                printf("%d %f %d\n",
                        1000*i,
                        (float)(j+1)/(num_threads+1),
                        final_num_recovereds[j]);
            }
        }
    }

    /* free memory for array */
    free(final_num_recovereds);

    /* tear down mpi */
    tryMPI(
            MPI_Finalize(),
            "MPI_Finalize",
            mpi_rank
          );

    /* quit program */
    return 0;
}
